#include <allegro.h>
#include <stdlib.h>

#include <iostream>
#include <vector>

using namespace std;

#include "data.h"
#include "common.h"
#include "classes.h"
#include "map.h"
#include "manager.h"
#include "enemies.h"
#include "playerinfo.h"
#include "ui.h"

DATAFILE *datafile;

// These bitmaps will all be the same size, but collisionmap
// Is in the lowest colour depth (i.e. 8)
BITMAP *background = 0;
BITMAP *collisionmap = 0;
BITMAP *vscreen = 0;
int backgroundWidth;
int backgroundHeight;

void ShowError(const char *msg)
{
	cerr << msg << endl;
}

int randombetween(int min, int max)
{
	return ((rand() % ((max - min) + 1)) + min);
}

static bool LoadData()
{
	set_color_conversion(COLORCONV_32_TO_16);
	datafile = load_datafile("data.dat");
	if (! datafile) {
		ShowError("Failed to load data.");
		return false;
	}
	return true;
}

void play_sound(int soundid, int speed, int vol)
{
	SAMPLE *spl = (SAMPLE *) datafile[soundid].dat;
	play_sample(spl, vol, 128, speed, 0);
}

// TODO: make it into a proper editor.
static void Editor()
{
	int x = 0, y = 0;
	int maxx, maxy;
	maxx = backgroundWidth - 640;
	maxy = backgroundHeight - 480;
	refreshvscreen();
	drawall();
	GameObject * lastobj = 0;
	show_mouse(screen);
	do {
		bool keyactive = false;
		for (int i=0; i<KEY_MAX; i++)
			keyactive = keyactive || key[i];

		if (mouse_b & 1) {
			lastobj = findnearest( x + mouse_x, y + mouse_y);
			keyactive = true;
		}
		if (keyactive) {
			show_mouse(0); // Hide mouse
			if (! key[KEY_LSHIFT]) {
				if ((key[KEY_LEFT]) && x>0) {
					x -= 4;
				}
				if ((key[KEY_RIGHT]) && x<maxx) {
					x += 4;
				}
				if ((key[KEY_UP]) && y>0) {
					y -= 4;
				}
				if ((key[KEY_DOWN]) && y< maxy) {
					y += 4;
				}
			} else {
				if (lastobj) {
					if (key[KEY_LEFT])
						lastobj->position.x --;
					if (key[KEY_RIGHT])
						lastobj->position.x ++;
					if (key[KEY_UP])
						lastobj->position.y --;
					if (key[KEY_DOWN])
						lastobj->position.y ++;
					StaticObject *so = dynamic_cast <StaticObject *> (lastobj);
					if (so) {
						so ->drawn = false;
					}
				}
			}
			if (keypressed()) {
				int key = readkey() & 0xff;
				// TODO: something.
				GameObject * obj = 0;
				if (key == '1') {
					obj = new StaticObject(TankHydrogen);
				}
				if (key == '2') {
					obj = new StaticObject(TankOxygen);
				}
				if (key == '3') {
					obj = new StaticObject(TankHCl);
				}
				if (key == '4') {
					obj = new StaticObject(TankHF);
				}
				if (key == '5') {
					obj = new StaticObject(PlayerBaseExit);
				}
				if (key == '%') {
					obj = new StaticObject(PlayerBaseStart);
				}
				if (key == '6') {
					obj = new Enemy(BlueStaticJelly);
				}
				if (key == '7') {
					obj = new Enemy(OrangeJelly);
				}
				if (key == '8') {
					obj = new Enemy(YellowJelly);
				}
				if (key == '9') {
					obj = new Enemy(GreenJelly);
				}
				if (key == '0') {
					// TODO: Find previous ship, get rid of it.
					obj = new Ship();
				}
				if (key == 'd') {
					if (lastobj)
					{
						lastobj->dead = true;
						lastobj = 0;
					}
				}
				if (obj) {
					obj->position.x = x + mouse_x;
					obj->position.y = y + mouse_y;
					addobject(obj);
					lastobj = obj;
				}
			}
			screenoffsetx = x;
			screenoffsety = y;
			// This is a kludge so the culling doesn't cull everything.
			viewportcentre.x = (float) x;
			viewportcentre.y = (float) y;
			removedeads();
			if (lastobj) {
				int ix, iy;
				ix = (int) lastobj->position.x;
				iy = (int) lastobj->position.y;
				circle(vscreen, ix,
						iy,
						48, makecol(255,255,255));
				addrect(ix - 50, iy - 50, 100, 100);
			}
			drawall();
			show_mouse(screen);
		}
	} while (! key[KEY_ESC]);
	saveobjects("leveledited.txt");
}

static void PlayGame(int startlevel)
{
	playerinfo.maxstartlevel = startlevel;
	while (uimainmenu())
	{
		playerinfo.init();
		do {
			playerinfo.levelcomplete = false;
			if (!uistartlevel()) {
				break;
			}
			initlevel(playerinfo.level);
			mainloop();
			if (playerinfo.levelcomplete)
			{
				uilevelcomplete();
				playerinfo.level++;
			}
		} while (playerinfo.levelcomplete && (playerinfo.level < NumLevels));
		// Game over. TODO: display game over message.
	}
}

int main(int argc, char *argv[])
{
	allegro_init();
	install_keyboard();
	install_timer();
	install_mouse();
	install_sound(DIGI_AUTODETECT, MIDI_NONE, 0);
	set_color_depth(16);
	int gfx_mode = GFX_AUTODETECT_FULLSCREEN;
	bool editor = false;
	int levelnum = 1; // Temporary
	if (argc >1) {
		if (argv[1][0] == 'e')
			editor = true;
	}
	if (argc >2) {
		levelnum = atoi(argv[2]);
	}
	if (editor) gfx_mode = GFX_AUTODETECT_WINDOWED;
	int res =
		set_gfx_mode(gfx_mode,640,480,640,480);
	if (res != 0) {
		ShowError("Cannot set gfx mode");
		return 1;
	}

	LoadData();
	initui();
	try {
		if (editor)
		{
			initlevel(levelnum);
			Editor();
		}
		else
			PlayGame(levelnum);
	} catch (const char *msg) {
		ShowError(msg);
		clear_keybuf();
		set_gfx_mode(GFX_TEXT,0,0,0,0);
		return 2;
	} catch (...) {
		ShowError("Unknown exception");
		return 3;
	}
	clear_keybuf();
	set_gfx_mode(GFX_TEXT,0,0,0,0);
	return 0;
}

END_OF_MAIN()

