#include <vector>
#include <allegro.h>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#include "common.h"
#include "classes.h"
#include "map.h"
#include "data.h"
#include "manager.h"
#include "enemies.h"

static const int MaxPoints = 20;

// Util function.
static void WideLine(BITMAP *bmp, int x1, int y1, int x2, int y2, int width, int colour)
{
	float dx = x2 - x1;
	float dy = y2 - y1;
	float len = sqrt(dx*dx + dy*dy);
	dx /= len;
	dy /= len;
	dx *= ((float) width) /2;
	dy *= ((float) width) /2;
	int offsx = (int) -dy;
	int offsy = (int) dx;
	int points[8];
	points[0] = x1 - offsx;	points[1] = y1 - offsy;
	points[2] = x1 + offsx;	points[3] = y1 + offsy;
	points[4] = x2 + offsx;	points[5] = y2 + offsy;
	points[6] = x2 - offsx;	points[7] = y2 - offsy;
	polygon(bmp, 4, points, colour);
}

static float lightx = -0.7;
static float lighty = 0.7;

static void LitLine(BITMAP *bmp, int x1, int y1, int x2, int y2)
{
	float dx = x2 - x1;
	float dy = y2 - y1;
	float len = sqrt(dx*dx + dy*dy);
	dx /= len;
	dy /= len;
	float result = (dx * lightx) + (dy * lighty);
	result = result + 1; // make sure it's positive.
	result = result * 128;
	int brightness = (int) result;
	if (brightness<0) brightness = 0;
	if (brightness>255) brightness = 255;
	int colour = makecol(brightness, brightness, brightness);
	for (int width=4; width <=12; width += 2)
		WideLine(bmp, x1, y1, x2, y2, width, colour);
}

Map::Map()
{

}

void Map::load(const char *filename)
{
	// Create input stream object
	ifstream infile(filename);
	if (! infile.good()) {
		throw "Failed to load map";
	}
	// Read the number of polys
	int polys;
	infile >> polys;
	// Resize the vector to hold them
	polygons.resize(polys);
	for (int i=0; i<polys; i++) {
		int points;
		// Read num points.
		infile >> points;
		if (points > MaxPoints) {
			throw "Too many points";
		}
		polygons[i].points.resize(points);
		for (int j=0; j<points; j++) {
			infile >> polygons[i].points[j].x;
			infile >> polygons[i].points[j].y;
		}
	}
	// All done.
}
/*
	drawing_mode(DRAW_MODE_TRANS,0,0,0);
	set_trans_blender(0,0,0,40);
	for (int i=0; i<10; i+= 2) {
		rectfill(background,100-i,100,200,200+i,makecol(255,255,255));
		rectfill(background,100,100-i,200+i,200,makecol(0,0,0));
	}
	drawing_mode(DRAW_MODE_SOLID,0,0,0);
	rectfill(background,100,100,200,200,makecol(0,0,0));
*/

void Map::drawcave()
{
	int pointarray[MaxPoints * 2];
	int polys = polygons.size();
	// Draw lines
	drawing_mode(DRAW_MODE_TRANS,0,0,0);
	set_trans_blender(0,0,0,40);
	for (int i=0; i<polys; i++) {
		int points = polygons[i].points.size();
		for (int j=0; j<(points - 1); j++) {
			LitLine(background,
				(int) polygons[i].points[j].x, (int) polygons[i].points[j].y,
				(int) polygons[i].points[j+1].x, (int) polygons[i].points[j+1].y);
		}
		LitLine(background,
			(int) polygons[i].points[points-1].x, (int) polygons[i].points[points-1].y,
			(int) polygons[i].points[0].x, (int) polygons[i].points[0].y);
	}
	drawing_mode(DRAW_MODE_SOLID,0,0,0);
	// Now hollow out the cave
	// And do the collision map.
	clear_to_color(collisionmap, 1);
	for (int i=0; i<polys; i++) {
		int points = polygons[i].points.size();
		for (int j=0; j<points; j++) {
			pointarray[j*2] = (int) polygons[i].points[j].x;
			pointarray[j*2 + 1] = (int) polygons[i].points[j].y;
		}
		polygon(background, points, pointarray, makecol(0,0,0));
		polygon(collisionmap, points, pointarray, makecol(0,0,0));
	}

}

// Background rendering functions

static void ApplyDirt(BITMAP *dirt, int qty)
{
	BITMAP * tempbitmaps[4];
	int w,h;
	w = dirt->w;
	h = dirt->h;
	// Create 4 bitmaps, flipped on various axes
	for (int i=0; i<4; i++) {
		tempbitmaps[i] = create_bitmap_ex(32,w,h);
		for (int x=0; x<w; x++)
		for (int y=0; y<h; y++)
		{
			int colour;
			colour = getpixel(dirt, x,y);
			int targetx = x;
			int targety = y;
			if ((i % 2) != 0)
				targetx = w-x;
			if ((i % 4) != 0)
				targety = h-y;
			putpixel(tempbitmaps[i],targetx,targety,colour);
		}
	}
	for (int i=0; i<qty; i++) {
		int x,y;
		x = randombetween(0, backgroundWidth-1);
		y = randombetween(0, backgroundHeight-1);
		x -= 128;
		y -= 128;
		int dir = randombetween(0,3);
		draw_trans_sprite(background, tempbitmaps[dir], x,y);
	}
	for (int i=0; i<4; i++)
		destroy_bitmap(tempbitmaps[i]);
}

static void InitBackgroundTexture()
{
	BITMAP * tile = (BITMAP *) datafile[REDTILE_TGA].dat;
	int dirtqty = ((backgroundWidth / 256) * (backgroundHeight / 256)) * 8;
	for (int x=0; x<backgroundWidth; x+= 128)
		for (int y=0; y<backgroundHeight; y+= 128)
			blit(tile, background, 0,0,x,y,256,256);
	set_alpha_blender();
	BITMAP * dirt = (BITMAP *) datafile[DIRT_TGA].dat;
	ApplyDirt(dirt, dirtqty);
	dirt = (BITMAP *) datafile[DIRT2_TGA].dat;
	ApplyDirt(dirt, dirtqty / 2);
	dirt = (BITMAP *) datafile[CRACK_TGA].dat;
	ApplyDirt(dirt, dirtqty / 2);
}

// Init the background and other bitmaps;
static void InitBackground()
{
	if (background != 0) {
		destroy_bitmap(background);
	}
	if (vscreen != 0) {
		destroy_bitmap(vscreen);
	}
	if (collisionmap != 0) {
		destroy_bitmap(collisionmap);
	}
	backgroundWidth = 2560;
	backgroundHeight = 2560;
	background = create_bitmap(backgroundWidth, backgroundHeight);
	vscreen = create_bitmap(backgroundWidth, backgroundHeight);
	collisionmap = create_bitmap_ex(8,backgroundWidth, backgroundHeight);
	if ((background == 0) || (vscreen == 0) || (collisionmap == 0)) {
		ShowError("Problem creating bitmaps (out of memory?)");
		throw "Problem creating bitmaps";
	}
}

static void InitObjects(const char *filename)
{
	ifstream in(filename);
	if (! in.good())
		throw "Failed to open objects file";
	do
	{
		char c;
		in >> c;
		if (in.eof()) {
			break;
		}
		GameObject *obj = 0;
		cout << "Object type: " << c << endl;
		switch(c) {
			case 'S': // Ship
				obj = new Ship();
				trackobject(obj);
				break;
			case 'T':
				{
					int statictype;
					in >> statictype;
					obj = new StaticObject((StaticType) statictype);
					break;
				}
			case 'E':
				{
					int enemytype;
					in >> enemytype;
					obj = new Enemy((EnemyType) enemytype);
					break;
				}
			default:
				throw "Unknown object type";
		}
		if (obj != 0) {
			in >> obj->position.x;
			in >> obj->position.y;
			// Put it in the collision map if necessary
			StaticObject * so = dynamic_cast <StaticObject *> (obj);
			if (so != 0) {
				so->drawtocollisionmap();
			}
			// add
			addobject(obj);
		}
	} while(true);
}


void initlevel(int level)
{
	// Clean any crap out of the manager.
	initmanager();
	// Create the background (or delete it and re-create)
	InitBackground();
	InitBackgroundTexture();
	ostringstream oss;
	oss << "levels/level" << level << ".txt";
	Map levelmap;
	levelmap.load(oss.str().c_str());
	levelmap.drawcave();
	// Load object data.
	ostringstream objectsfilename;
	objectsfilename << "levels/level" << level << "objects.txt";
	InitObjects(objectsfilename.str().c_str());
	// DEBUG!
	/*
	blit(collisionmap, screen, 400,0,0,0,640,480);
	rest(5000);
	*/

}
